""" Code for rendering a D3-based graph in Jupyter.
@author Michael Stewart
"""

import pathlib
import os
import random

import neo4j
from neo4j import GraphDatabase
from neo4j.exceptions import AuthError
from IPython.core.display import display, HTML
from string import Template

NEO4J_HOST = "neo4j://localhost:7687"


def visualise_list(nodes: list, links: list):
    """Visualise the given list of nodes and links via a D3 graph.

    Args:
        nodes (list): The list of nodes.
        links (list): The list of links (edges).

    Returns:
        HTML: A HTML snippet with the graph rendered inside of it.

    Raises:
        ValueError: If any arguments are no good.
    """
    if type(nodes) != list:
        raise ValueError(
            "Illegal argument: The first argument must be a list of nodes."
        )
    if len(nodes) == 0:
        raise ValueError(
            "Illegal argument: nodes must contain at least "
            "one node to render."
        )
    if type(links) != list:
        raise ValueError(
            "Illegal argument: The second argument must be a " "list of links."
        )

    template = _load_template()

    return _build_html_template(nodes, links)


def visualise_cypher(query: str):
    """Visualise the given cypher query via a D3 graph.

    Args:
        query (str): The Cypher query to visualise.

    Returns:
        HTML: A HTML snippet with the graph rendered inside of it.

    Raises:
        ValueError: If connect_to_neo4j has not yet been called.
    """

    graph_driver = _connect_to_neo4j()

    # Run the query via neo4j.
    with graph_driver.session() as session:
        nodes, links = session.read_transaction(_run_neoj4_query, query)

    return _build_html_template(nodes, links)


def _connect_to_neo4j():
    """Connect to Neo4j so that cypher queries can be visualised.
    At the moment, only connects to the default port (feel free to change
    if you need to run on docker or something.)

    Args:
        graph_password (str, optional): The password of the neo4j db.
    """
    graph_password: str = "password"
    if "NEO4J_PASSWORD" in os.environ:
        graph_password = os.environ["NEO4J_PASSWORD"]
    try:
        graph_driver = GraphDatabase.driver(
            NEO4J_HOST, auth=("neo4j", graph_password)
        )
        return graph_driver
    except AuthError as e:
        raise ValueError(
            f"Could not connect to Neo4j as the password "
            f"'{graph_password}' is not correct."
        )
    except Exception as e:
        raise ValueError(
            "Unable to connect to Neo4j. Please ensure it is running."
        )


def _build_html_template(nodes: list, links: list):
    """Return a HTML template of the given nodes and links.

    Args:
        nodes (list): The list of nodes.
        links (list): The list of links (edges).

    Returns:
        HTML: An injected HTML template.
    """
    template = _load_template()
    return HTML(
        Template(template).safe_substitute(
            {
                # Generate a random chart id so that D3 knows which svg
                # element corresponds to this graph.
                "chart_id": str(random.randint(0, 1000000)),
                "nodes": nodes,
                "links": links,
            }
        )
    )


def _load_template():
    """Load the template from template.html. This will be populated with the
    given nodes and links later.
    """
    current_path = pathlib.Path(__file__).parent.resolve()
    template = ""
    with open(
        os.path.join(current_path, "template", "template.html"), "r"
    ) as f:
        template = f.read()
    return template


def _run_neoj4_query(tx, query):
    """Run the given query via Neo4j. Return a list of nodes, and a list of
    links (edges), that can be visualised via D3Graph's visualise() method.

    Args:
        tx (neo4j): The neo4j driver.
        query (str): The cypher query to run.

    Returns:
        list, list: A list of nodes, and list of links.
    """
    result = tx.run(query)
    result_ls = []

    nodes_dict = {}
    links = []

    for record in result:

        for (k, v) in record.items():
            if type(v) != neo4j.graph.Node:
                links.append(
                    {
                        "source": v.nodes[0].id,
                        "target": v.nodes[1].id,
                        "type": v.type,
                        **dict(v),
                    }
                )
            else:
                # Ignore non str and int for now.
                # Datetimes are a bit weird to parse into JS,
                # so any dates will not be included in the properties
                # in the visualisation.
                nodes_dict[v.id] = {
                    "id": v.id,
                    "category": list(v.labels)[-1]  # 'category' corresponds
                    # to the final label of the entity, and is
                    # responsible for the node colouring.
                    if len(v.labels) > 0 else None,
                    **dict(
                        (
                            k2,
                            v2.replace("'", "")  # Replace single quotes
                            # with double quotes so they can be
                            # rendered in JSON
                            if type(v2) == str else v2,
                        )
                        for (k2, v2) in v.items()
                        if type(v2) in [str, int]
                    ),
                }
    # print(links)
    # print(nodes_dict.values())
    return list(nodes_dict.values()), links



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gqvis']

package_data = \
{'': ['*'], 'gqvis': ['template/*']}

install_requires = \
['ipython==7.34.0', 'neo4j>=5.2.0,<6.0.0']

setup_kwargs = {
    'name': 'gqvis',
    'version': '0.1.10',
    'description': 'A simple package for visualising the results of a cypher-based graph query to Neo4j in Python.',
    'long_description': '# GQVis\n\nA simple package for visualising the results of a graph query (to Neo4j) in an interactive Python environment (such as Jupyter notebook).\n\n![Screenshot of an example graph](https://github.com/nlp-tlp/gqvis/blob/main/image_1.png?raw=true)\n\n## Installation\n\nSimply run:\n\n    pip install gqvis\n\n## Usage\n\nFirst, import the package via\n\n    import gqvis\n\nThere are two ways to use this package, detailed below.\n\n### Visualising the result of a Neo4j Cypher query\n\nThe first way to use the class is to visualise the result of a Neo4j Cypher query. This requires you to have a Neo4j database running.\n\nYou must first create an environment variable called `NEO4J_PASSWORD` and set the value of this variable to the password of your Neo4j graph. In Jupyter, you can simply do:\n\n    %env NEO4J_PASSWORD=password\n\nThe default password is \'password\', so if your Neo4j graph already has this password there is no need to set this variable.\n\nThen, you can run the following:\n\n    gqvis.visualise_cypher(\'MATCH (n1:Entity)-[r]->(n2:Entity) RETURN n1, r, n2 LIMIT 500\')\n\nNote that unlike Neo4j, which has the \'connect result nodes\' option to automatically connect nodes that have relationships, you will need to return the relationships explicitly in your query. Only relationships in the `RETURN` statement will be visualised.\n\n### Visualising nodes and links directly\n\nYou can also visualise a given list of nodes and edges - No Neo4j required. For example:\n\n    nodes = [\n        {\n          "id": 1,\n          "category": "Person",\n          "name": "Bob",\n        },\n        {\n          "id": 2,\n          "category": "Food",\n          "name": "Jelly",\n        },\n        {\n          "id": 3,\n          "category": "Person",\n          "name": "Alice",\n        }\n    ]\n    links = [\n        {\n          "source": 1,\n          "target": 2,\n          "type": "EATS",\n        },\n        {\n          "source": 3,\n          "target": 1,\n          "type": "LIKES",\n        },\n    ]\n    gqvis.visualise_list(nodes, links)\n\nThis will create a graph visualisation with three nodes ("Bob", "Jelly", "Alice"), and two links (Bob eats Jelly, Alice likes Bob). You can have other properties (such as `"age": 45` on Bob) - they\'ll be shown in the tooltip when hovering over a node.\n\nThe `"id"`, `"category"` and `"name"` properties are required on each node. The `"name"` property is what will be written on the nodes in the visualisation, while the `"category"` will determine their colour (more on this below).\n\nFor the links, `"source"` is the id of the source node, `"target"` is the id of the target node, and `"type"` is the type of relationship. These are all required.\n\n### About the visualisation\n\nNodes are coloured based on the `category` property.\n\nFor the Cypher visualisation, the way the graph decides on the colour of each node is based on the last label of that node, i.e. if a node had the following labels:\n\n    Entity, Item\n\n... it would be coloured based on the `Item` label. The colours are determined automatically, i.e. each category receives its own unique colour.\n\n## Notes\n\nWe use the dependency `neo4j` (the Neo4j driver for Python) rather than `py2neo` because it turns out `py2neo` does not output the exact same results as Neo4j. The way this whole thing works is by creating a list of nodes from all node objects returned by the cypher query, then creating links (by linking nodes via their ids). It didn\'t seem possible in `py2neo`, but was pretty straightforward with the `neo4j` package.\n\nYou can run `src/template/template.html` by itself (i.e. open it directly in Firefox/Chrome) for development purposes. When running it this way, it will be populated by some dummy data specified in `src/template/dummyData.js`.\n\nNote that you must have an internet connection to use this package at the moment as it pulls D3.js from an online CDN.\n',
    'author': 'Michael Stewart',
    'author_email': 'michael.stewart.webdev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
